/*
 * NJS module for Python.
 *
 * Exposes the NJS API to Python.
 * Tested with version NJS 0.3.0.
 *
 * David Leonard, 2001. Public domain.
 *
 * TODO: NJS may be multithreaded, so need to release/acquire interpreter
 * lock on calls into njs.
 *
 * $Id: njsmodule.c,v 1.2 2001/11/24 01:59:11 d Exp $
 */

#define VERSION		"$Revision: 1.2 $"

#include "Python.h"
#include "njs/njs.h"

/* Unique objects */
static PyObject *NjsErrorObject;
static PyObject *NjsImmutableErrorObject;
static PyObject *NjsPredefinedErrorObject;
static PyObject *NjsTrueObject;
static PyObject *NjsFalseObject;
static PyObject *NjsUndefinedObject;

/* NJS Class used to wrap untranslatable Python objects */
static NJSClassPtr NjsPyClass;

/* Debugging macro */
#if 1
#include <stdio.h>
#define X(m, args...)	fprintf(stderr, m "\n" , ## args); 
#else
#define X(m, args...)
#endif

/* Some early declarations to get around later usage */
typedef struct {
	PyObject_HEAD
	NJSInterpPtr	interp;
	int		destroy_on_deref;
} NjsInterpObject;

static NjsInterpObject *newNjsInterpObject(NJSInterpPtr, int);
staticforward PyTypeObject NjsInterp_Type;
#define NjsInterpObject_Check(v)	((v)->ob_type == &NjsInterp_Type)

/* Prototypes */

static void	 njserror(NJSInterpPtr, char *);
static PyObject *NjsError(NJSInterpPtr, const char *);
static void	 decref_callback(void *);
static PyObject *interp_result(NjsInterpObject *);

/*------------------------------------------------------------
 * Error handling
 */

/* Translate a Python error into an error for the NJS library */
static void
njserror(NJSInterpPtr interp, char *error_return)
{
	if (!PyErr_Occurred())
		snprintf(error_return, NJS_ERROR_RETURN_SIZE, "python: no error?");
	else {
		PyObject *t, *v, *tb;
		PyObject *ts = NULL, *vs = NULL;
		char *tss = NULL, *vss = NULL;
		NJSInterpOptions opt;

		/* Be careful here as new errors will overwrite the original. */
		PyErr_Fetch(&t, &v, &tb);
		Py_XINCREF(t);
		Py_XINCREF(v);
		Py_XINCREF(tb);

		/* Carefully turn the exception objects into strings */
		if (t != NULL) {
			ts = PyObject_Str(t);
			if (ts != NULL) {
				tss = PyString_AsString(ts);
				Py_DECREF(ts);
			}
		}
		if (v != NULL && v != Py_None) {
			vs = PyObject_Str(v);
			if (ts != NULL) {
				vss = PyString_AsString(vs);
				Py_DECREF(vs);
			}
		}

		/* Construct a string describing the exception for NJS */
		snprintf(error_return, NJS_ERROR_RETURN_SIZE,
			"%s%s%s", tss ? tss : "(null)", vss ? ": " : "", vss ? vss :  "");

		/* Dump stacktrace to sys.stderr if NJS is also doing tracebacks */
		njs_get_options(interp, &opt);
		if (opt.stacktrace_on_error) {
			/* Restore original exception in case it changed */
			PyErr_Restore(t, v, tb);
			PyErr_Print();
		}

		Py_XDECREF(t);
		Py_XDECREF(v);
		Py_XDECREF(tb);

		/* Signal to Python that we have handled the exception. */
		PyErr_Clear();
	}
}

/* Translate an NJS library error into a Python exception */
static PyObject *
NjsError(NJSInterpPtr interp, const char *msg)
{
	char *errmsg = njs_error_message(interp);

	PyErr_SetObject(NjsErrorObject, 
		Py_BuildValue(errmsg ? "zz" : "z", msg, errmsg));
	return NULL;
}

/*------------------------------------------------------------
 * Booleans
 */

typedef struct {
	PyObject_HEAD
	int	value;
} NjsBooleanObject;

staticforward PyTypeObject NjsBoolean_Type;

#define NjsBooleanObject_Check(v)	((v)->ob_type == &NjsBoolean_Type)

/* Construct a boolean object */
static NjsBooleanObject *
newNjsBooleanObject(int value)
{
	NjsBooleanObject *self;

	self = PyObject_New(NjsBooleanObject, &NjsBoolean_Type);
	if (self == NULL)
		return NULL;
	self->value = value;
	return self;
}

/* Free a boolean object */
static void
NjsBoolean_dealloc(NjsBooleanObject *self)
{
	PyObject_Del(self);
}

/* Return true if the boolean is true */
static int
NjsBoolean_nonzero(NjsBooleanObject *self)
{
	return self->value;
}

/* Return a string representation of the boolean */
static PyObject *
NjsBoolean_repr(NjsBooleanObject *self)
{
	return PyString_FromString(
		self->value ? "<njs.true>"
			    : "<njs.false>");
}

/* Number protocol, but we only use the nonzero method (for 'if' statements) */
static PyNumberMethods NjsBoolean_Number = {
        0,		/* nb_add */
        0,		/* nb_subtract */
        0,		/* nb_multiply */
        0,		/* nb_divide */
        0,		/* nb_remainder */
        0,		/* nb_divmod */
        0,		/* nb_power */
        0,		/* nb_negative */
        0,		/* nb_positive */
        0,		/* nb_absolute */
        (inquiry)NjsBoolean_nonzero,		/* nb_nonzero */
};

statichere PyTypeObject NjsBoolean_Type = {
	PyObject_HEAD_INIT(NULL)
	0,			/*ob_size*/
	"NJSBoolean",		/*tp_name*/
	sizeof(NjsBooleanObject),/*tp_basicsize*/
	0,			/*tp_itemsize*/
	/* methods */
	(destructor)NjsBoolean_dealloc,	/*tp_dealloc*/
	0,			/*tp_print*/
	0,			/*tp_getattr*/
	0,			/*tp_setattr*/
	0,			/*tp_compare*/
	(reprfunc)NjsBoolean_repr,/*tp_repr*/
	&NjsBoolean_Number,	/*tp_as_number*/
	0,			/*tp_as_sequence*/
	0,			/*tp_as_mapping*/
	0,			/*tp_hash*/
};



/*------------------------------------------------------------
 * Undefined
 */

staticforward PyTypeObject NjsUndefined_Type;

#define NjsUndefinedObject_Check(v)	((v)->ob_type == &NjsUndefined_Type)

/* Allocate storage for the undefined object */
static PyObject *
newNjsUndefinedObject(void)
{
	return PyObject_New(PyObject, &NjsUndefined_Type);
}

/* Release storage for the undefined object */
static void
NjsUndefined_dealloc(PyObject *self)
{
	PyObject_Del(self);
}

/* Always return false equivalence for undefined */
static int
NjsUndefined_nonzero(PyObject *self)
{
	return 0;
}

/* Return a string representation for the undefined object */
static PyObject *
NjsUndefined_repr(PyObject *self)
{
	return PyString_FromString("<njs.undefined>");
}

/* Number protocol, only using nonzero */
static PyNumberMethods NjsUndefined_Number = {
        0,		/* nb_add */
        0,		/* nb_subtract */
        0,		/* nb_multiply */
        0,		/* nb_divide */
        0,		/* nb_remainder */
        0,		/* nb_divmod */
        0,		/* nb_power */
        0,		/* nb_negative */
        0,		/* nb_positive */
        0,		/* nb_absolute */
        (inquiry)NjsUndefined_nonzero,		/* nb_nonzero */
};

statichere PyTypeObject NjsUndefined_Type = {
	PyObject_HEAD_INIT(NULL)
	0,			/*ob_size*/
	"NJSUndefined",		/*tp_name*/
	sizeof(PyObject),	/*tp_basicsize*/
	0,			/*tp_itemsize*/
	/* methods */
	(destructor)NjsUndefined_dealloc,/*tp_dealloc*/
	0,			/*tp_print*/
	0,			/*tp_getattr*/
	0,			/*tp_setattr*/
	0,			/*tp_compare*/
	(reprfunc)NjsUndefined_repr,/*tp_repr*/
	&NjsUndefined_Number,	/*tp_as_number*/
	0,			/*tp_as_sequence*/
	0,			/*tp_as_mapping*/
	0,			/*tp_hash*/
};

/*------------------------------------------------------------
 * NJSVMDispatchMethod
 */

typedef struct {
	PyObject_HEAD
	NJSVMDispatchMethod value;
} NjsVMDispatchMethodObject;

staticforward PyTypeObject NjsVMDispatchMethod_Type;

#define NjsVMDispatchMethodObject_Check(v)	\
			((v)->ob_type == &NjsVMDispatchMethod_Type)

/* Return a new or existing enumerated NJSVMDispatchMethod value */
static NjsVMDispatchMethodObject *
newNjsVMDispatchMethodObject(NJSVMDispatchMethod value)
{
	/* A cache of existing VMDispatchMethod values to save memory */
	static NjsVMDispatchMethodObject *cache[4];

	NjsVMDispatchMethodObject *self;

	if ((int)value < 0 || (int)value > sizeof cache / sizeof cache[0]) {
		PyErr_SetNone(PyExc_ValueError);
		return NULL;
	}

	if ((self = cache[(int)value]) != NULL) {
		Py_INCREF(self);
		return self;
	}

	self = PyObject_New(NjsVMDispatchMethodObject,
			&NjsVMDispatchMethod_Type);
	if (self == NULL)
		return NULL;
	self->value = value;

	/* Keep a copy in the cache for later */
	cache[(int)value] = self;
	Py_INCREF(self);

	return self;
}

/* Return a string representation of NJSVMDispatchMethod enumerations */
static PyObject *
NjsVMDispatchMethod_repr(NjsVMDispatchMethodObject *self)
{
	char *name;
	switch (self->value) {
	case NJS_VM_DISPATCH_SWITCH_BASIC:
		name = "<njs.VM_DISPATCH_SWITCH_BASIC>"; break;
	case NJS_VM_DISPATCH_SWITCH:
		name = "<njs.VM_DISPATCH_SWITCH>"; break;
	case NJS_VM_DISPATCH_JUMPS:
		name = "<njs.VM_DISPATCH_JUMPS>"; break;
	default:
		PyErr_SetString(PyExc_SystemError, 
			"NjsVMDispatchMethod_repr");
		return NULL;
	}
	return PyString_FromString(name);
}

/* Release storage for NJSVMDispatchMethod objects */
static void
NjsVMDispatchMethod_dealloc(NjsVMDispatchMethodObject *self)
{
	PyObject_Del(self);
}

statichere PyTypeObject NjsVMDispatchMethod_Type = {
	PyObject_HEAD_INIT(NULL)
	0,			/*ob_size*/
	"VMDispatchMethod",	/*tp_name*/
	sizeof(NjsVMDispatchMethodObject),/*tp_basicsize*/
	0,			/*tp_itemsize*/
	/* methods */
	(destructor)NjsVMDispatchMethod_dealloc,/*tp_dealloc*/
	0,			/*tp_print*/
	0,			/*tp_getattr*/
	0,			/*tp_setattr*/
	0,			/*tp_compare*/
	(reprfunc)NjsVMDispatchMethod_repr,/*tp_repr*/
	0,			/*tp_as_number*/
	0,			/*tp_as_sequence*/
	0,			/*tp_as_mapping*/
	0,			/*tp_hash*/
};

/*------------------------------------------------------------
 * NJSInterpOptions
 */

typedef struct {
	PyObject_HEAD
	NJSInterpOptions value;
	struct iofcontext {
		PyObject	*file;
		int		writep;		/* 0 = write, 1 = read */
	} in, out, err;
	PyObject	*hookfn;
} NjsInterpOptionsObject;

staticforward PyTypeObject NjsInterpOptions_Type;

#define NjsInterpOptionsObject_Check(v)	\
			((v)->ob_type == &NjsInterpOptions_Type)

/* Construct a new (uninitialised) NJSInterpOptions object */
static NjsInterpOptionsObject *
newNjsInterpOptionsObject(void)
{
	NjsInterpOptionsObject *self;

	self = PyObject_New(NjsInterpOptionsObject,
			&NjsInterpOptions_Type);
	if (self == NULL)
		return NULL;

	self->in.file = NULL;
	self->in.writep = 0;
	self->out.file = NULL;
	self->out.writep = 1;
	self->err.file = NULL;
	self->err.writep = 1;

	self->hookfn = NULL;

	return self;
}

/* Free storage associated with an InterpOptions object */
static void
NjsInterpOptions_dealloc(NjsInterpOptionsObject *self)
{
	Py_XDECREF(self->in.file);
	Py_XDECREF(self->out.file);
	Py_XDECREF(self->err.file);
	Py_XDECREF(self->hookfn);
	PyObject_Del(self);
}

/* Retreive a particular option within the InterpOptions object */
static PyObject *
NjsInterpOptions_getattr(NjsInterpOptionsObject *self, char *name)
{
	PyObject *result = NULL;

	/* Convenient list of member attributes */
	if (strcmp(name, "__members__") == 0) {
		static char *members[] = {
			"annotate_assembler",
			"debug_info",
			"dispatch_method",
			"executable_bc_files",
			"fd_count",
			"hook",
			"hook_operand_count_trigger",
			"no_compiler",
			"only_define_ecma",
			"optimize_bc_size",
			"optimize_heavy",
			"optimize_jumps_to_jumps",
			"optimize_peephole",
			"s_stderr",
			"s_stdin",
			"s_stdout",
			"secure_builtin_file",
			"secure_builtin_system",
			"stack_size",
			"stacktrace_on_error",
			"verbose",
			"warn_deprecated",
			"warn_missing_semicolon",
			"warn_shadow",
			"warn_strict_ecma",
			"warn_undef",
			"warn_unused_argument",
			"warn_unused_variable",
			"warn_with_clobber",
		};
		int i;
		result = PyList_New(sizeof members / sizeof members[0]);
		for (i = 0; i < sizeof members / sizeof members[0]; i++)
		     PyList_SetItem(result, i, PyString_FromString(members[i]));
		return result;
	}

	/* All the integer attributes */
#define D(n) 						\
	if (strcmp(name, #n) == 0)			\
		return PyInt_FromLong(self->value.n);
	D(stack_size)
	D(verbose)
	D(no_compiler)
	D(only_define_ecma)
	D(stacktrace_on_error)
	D(secure_builtin_file)
	D(secure_builtin_system)
	D(annotate_assembler)
	D(debug_info)
	D(executable_bc_files)
	D(warn_unused_argument)
	D(warn_unused_variable)
	D(warn_undef)
	D(warn_shadow)
	D(warn_with_clobber)
	D(warn_missing_semicolon)
	D(warn_strict_ecma)
	D(warn_deprecated)
	D(optimize_peephole)
	D(optimize_jumps_to_jumps)
	D(optimize_bc_size)
	D(optimize_heavy)
	D(hook_operand_count_trigger)
	D(fd_count)
#undef D

	/* An enumerated type attribute */
	if (strcmp(name, "dispatch_method") == 0)
		return (PyObject *)newNjsVMDispatchMethodObject(
			self->value.dispatch_method);

	/* Object fields that could be NULL, may need to be turned into None */
	if (strcmp(name, "hook") == 0) 
		result = self->hookfn;
	else if (strcmp(name, "s_stdout") == 0)
		result = self->out.file;
	else if (strcmp(name, "s_stdin") == 0)
		result = self->in.file;
	else if (strcmp(name, "s_stderr") == 0) 
		result = self->err.file;
	else {
		/* All other attribute names are invalid */
		PyErr_SetString(PyExc_AttributeError, name);
		return NULL;
	}
	if (result == NULL)
		result = Py_None;
	Py_INCREF(result);
	return result;
}

/*
 * Generic IO callback function.
 * This is called when the s_std{in,out,err} field of the options has been
 * set. We know the direction of the IO from 'writep'.
 */
static int
iofunc_generic(PyObject *file, unsigned char *buf, unsigned int amt, int writep)
{
	PyObject *result;
	PyObject *buffer;

	if (file == NULL) 
		return -1;

	/* Create a temporary Python 'buffer' object */
	buffer = PyBuffer_FromReadWriteMemory(buf, amt);
	if (buffer == NULL) {
		PyErr_Clear();		/* XXX how to return error? */
		return -1;
	}

	/* Call the s_std{in,out,err}'s read() or write() method */
	result = PyObject_CallMethod(file, writep ? "write" : "read", 
		"iO", amt, buffer);

	Py_DECREF(buffer);
	if (result == NULL) {
		PyErr_Clear();		/* XXX how to return error? */
		return -1;
	}

	Py_DECREF(result);		/* XXX use the result for amt? */
	return amt;			/* XXX what should we return? */
}

/* Callback for stdin IO */
static int
iofunc_stdin(void *ctxt, unsigned char *buf, unsigned int amt)
{
	NjsInterpOptionsObject *self = (NjsInterpOptionsObject *)ctxt;
	return iofunc_generic(self->in.file, buf, amt, 0);
}

/* Callback for stdout IO */
static int
iofunc_stdout(void *ctxt, unsigned char *buf, unsigned int amt)
{
	NjsInterpOptionsObject *self = (NjsInterpOptionsObject *)ctxt;
	return iofunc_generic(self->out.file, buf, amt, 1);
}

/* Callback for stderr IO */
static int
iofunc_stderr(void *ctxt, unsigned char *buf, unsigned int amt)
{
	NjsInterpOptionsObject *self = (NjsInterpOptionsObject *)ctxt;
	return iofunc_generic(self->err.file, buf, amt, 1);
}

/*
 * Set up the option's NJSIOFunc to be either NULL or to call 
 * one of the previous three IO callbacks.
 */
static int
iofunc_set(PyObject **fp, NJSIOFunc* iofp, PyObject *v,
		NJSIOFunc iof, int writep)
{
	if (v == Py_None) {
		Py_XDECREF(*fp);
		*fp = NULL;
		*iofp = NULL;
		return 0;
	}
	/*
	 * XXX should we check here to see if v has 
	 * a callable read or write method?
	 */
	Py_XDECREF(*fp);
	*fp = v;
	Py_INCREF(v);
	*iofp = iof;
	return 0;
}

/* Set an InterpOptions attribute by name */
static int
NjsInterpOptions_setattr(NjsInterpOptionsObject *self, char *name, PyObject *v)
{
	long val;

	/* The IO redirectors */
	if (strcmp(name, "s_stdin") == 0)
		return iofunc_set(&self->in.file, &self->value.s_stdin, 
			v, iofunc_stdin, self->in.writep);
	if (strcmp(name, "s_stdout") == 0)
		return iofunc_set(&self->out.file, &self->value.s_stdout, 
			v, iofunc_stdout, self->out.writep);
	if (strcmp(name, "s_stderr") == 0)
		return iofunc_set(&self->err.file, &self->value.s_stderr, 
			v, iofunc_stderr, self->err.writep);

	/* Enumerated types */
	if (strcmp(name, "dispatch_method") == 0) {
		NjsVMDispatchMethodObject *mo = (NjsVMDispatchMethodObject*)v;
		if (!NjsVMDispatchMethodObject_Check(v)) {
			PyErr_SetString(PyExc_TypeError, name);
			return -1;
		}
		self->value.dispatch_method = mo->value;
		return 0;
	}

	/* XXX setting hook is not implemented */
	if (strcmp(name, "hook") == 0) {
		PyErr_SetNone(PyExc_NotImplementedError);
		return -1;
	}

	/* Everything else is integer */
	if (!PyInt_Check(v)) {
		PyErr_SetString(PyExc_TypeError, name);
		return -1;
	}
	val = PyInt_AsLong(v);

#define D(n) 						\
	if (strcmp(name, #n) == 0) {			\
		self->value.n = val;			\
		return 0;				\
	}
	D(stack_size)
	D(verbose)
	D(no_compiler)
	D(only_define_ecma)
	D(stacktrace_on_error)
	D(secure_builtin_file)
	D(secure_builtin_system)
	D(annotate_assembler)
	D(debug_info)
	D(executable_bc_files)
	D(warn_unused_argument)
	D(warn_unused_variable)
	D(warn_undef)
	D(warn_shadow)
	D(warn_with_clobber)
	D(warn_missing_semicolon)
	D(warn_strict_ecma)
	D(warn_deprecated)
	D(optimize_peephole)
	D(optimize_jumps_to_jumps)
	D(optimize_bc_size)
	D(optimize_heavy)
	D(hook_operand_count_trigger)
	D(fd_count)
#undef D

	/* Every other name is unknown */
	PyErr_SetString(PyExc_AttributeError, name);
	return -1;
}

statichere PyTypeObject NjsInterpOptions_Type = {
	PyObject_HEAD_INIT(NULL)
	0,			/*ob_size*/
	"InterpOptions",	/*tp_name*/
	sizeof(NjsInterpOptionsObject),	/*tp_basicsize*/
	0,			/*tp_itemsize*/
	/* methods */
	(destructor)NjsInterpOptions_dealloc,		/*tp_dealloc*/
	0,			/*tp_print*/
	(getattrfunc)NjsInterpOptions_getattr,		/*tp_getattr*/
	(setattrfunc)NjsInterpOptions_setattr,		/*tp_setattr*/
	0,			/*tp_compare*/
	0,			/*tp_repr*/
	0,			/*tp_as_number*/
	0,			/*tp_as_sequence*/
	0,			/*tp_as_mapping*/
	0,			/*tp_hash*/
};


/*------------------------------------------------------------
 * NJSValue (opaque)
 */

typedef struct {
	PyObject_HEAD
	NJSValue	value;
	NjsInterpObject	*interp;
} NjsValueObject;

staticforward PyTypeObject NjsValue_Type;

#define NjsValueObject_Check(v) ((v)->ob_type == &NjsValue_Type)

/* Allocate storage for an NJSValue wrapper object */
static NjsValueObject *
newNjsValueObject(NjsInterpObject *interp, NJSValue *value)
{
	NjsValueObject *self;

	if (value == NULL) {
		PyErr_SetString(PyExc_SystemError, "newNjsValueObject");
		return NULL;
	}

	self = PyObject_New(NjsValueObject, &NjsValue_Type);
	if (self == NULL)
		return NULL;

	/* Keep a copy of the value to stop garbage collection */
	memcpy(&self->value, value, sizeof self->value);
	njs_gc_pin(interp->interp, &self->value);

	self->interp = interp;
	Py_INCREF(interp);

	return self;
}

/* Release storage for a NJSValue object */
static void
NjsValue_dealloc(NjsValueObject *self)
{
	njs_gc_unpin(self->interp->interp, &self->value);
	Py_DECREF(self->interp);
	PyObject_Del(self);
}

/* Return a string representation of NJSVMDispatchMethod enumerations */
static PyObject *
NjsValue_repr(NjsValueObject *self)
{
	char buf[1024];

	snprintf(buf, sizeof buf, "<NJSValue at 0x%x type %d>", 
		(int)self, self->value.type);
	return PyString_FromString(buf);
}

statichere PyTypeObject NjsValue_Type = {
	/* The ob_type field must be initialized in the module init function
	 * to be portable to Windows without using C++. */
	PyObject_HEAD_INIT(NULL)
	0,			/*ob_size*/
	"NJSValue",		/*tp_name*/
	sizeof(NjsValueObject),	/*tp_basicsize*/
	0,			/*tp_itemsize*/
	/* methods */
	(destructor)NjsValue_dealloc, /*tp_dealloc*/
	0,			/*tp_print*/
	0,			/*tp_getattr*/
	0,			/*tp_setattr*/
	0,			/*tp_compare*/
	(reprfunc)NjsValue_repr,/*tp_repr*/
	0,			/*tp_as_number*/
	0,			/*tp_as_sequence*/
	0,			/*tp_as_mapping*/
	0,			/*tp_hash*/
};

/* Utility function to convert an NJSValue to a PyObject */
static PyObject *
NJSValue_to_PyObject(NjsInterpObject *interp, NJSValue *v)
{
	PyObject *res;
	int i;

	if (v == NULL) {
		PyErr_SetString(PyExc_SystemError, "NJSValue_to_PyObject");
		return NULL;
	}

	/* Convert the value into an equivalent Python object */
	switch (v->type) {
	case NJS_VALUE_UNDEFINED:
		Py_INCREF(NjsUndefinedObject);
		return NjsUndefinedObject;
	case NJS_VALUE_NULL:
		Py_INCREF(Py_None);
		return Py_None;
	case NJS_VALUE_BOOLEAN:
		res = v->u.i ? NjsTrueObject : NjsFalseObject;
		Py_INCREF(res);
		return res;
	case NJS_VALUE_INTEGER:
		return PyInt_FromLong(v->u.i);
	case NJS_VALUE_STRING:
		/* XXX Unicode? */
		return PyString_FromStringAndSize(v->u.s->data, v->u.s->len);
	case NJS_VALUE_DOUBLE:
		return PyFloat_FromDouble(v->u.d);
	case NJS_VALUE_ARRAY:
		/* XXX should be able to create a mutable array object */
		res = PyTuple_New(v->u.array->length);
		if (res == NULL)
			return NULL;
		for (i = 0; i < v->u.array->length; i++) {
			PyObject *o = NJSValue_to_PyObject(interp, 
				&v->u.array->data[i]);
			if (o == NULL) {
				Py_DECREF(res);
				return NULL;
			}
			PyTuple_SET_ITEM(res, i, o);
		}
		return res;
	default:
		/* Is it a NjsPyClass object */
		if (njs_isa(interp->interp, v, NjsPyClass, (void **)&res))
		{
			Py_INCREF(res);
			return res;
		}

		/* All other values we need to wrap opaquely */
		return (PyObject *)newNjsValueObject(interp, v);
	}
}

/* Utility function to convert argv of NJSValues into a Python tuple */
static PyObject *
NJSValues_to_PyObjects(NjsInterpObject *interp, int argc, NJSValue *argv)
{
	int i;
	PyObject *result;

	result = PyTuple_New(argc);
	for (i = 0; i < argc; i++)
		PyTuple_SET_ITEM(result, i, 
		     NJSValue_to_PyObject(interp, &argv[i]));
	return result;
}

/* Utility function to convert a python value into a NJSValue */
static int
PyObject_to_NJSValue(NJSInterpPtr interp, NJSValue *v, PyObject *obj,
	char *error_return)
{
	if (obj == NULL || NjsUndefinedObject_Check(obj)) {
		v->type = NJS_VALUE_UNDEFINED;
	}
	else if (obj == Py_None) {
		v->type = NJS_VALUE_NULL;
	}
	else if (NjsValueObject_Check(obj)) {
		/* Opaque value wrappers are simply unwrapped */
		NjsValueObject *vo = (NjsValueObject *)obj;
		memcpy(v, &vo->value, sizeof *v);
	}
	else if (NjsBooleanObject_Check(obj)) {
		v->type = NJS_VALUE_BOOLEAN;
		v->u.i = PyObject_IsTrue(obj);
	}
	else if (PyNumber_Check(obj)) {
		if (PyInt_Check(obj)) {
			v->type = NJS_VALUE_INTEGER;
			v->u.i = PyInt_AsLong(obj);
		} else {
			/* All other number formats we convert to float */
			PyObject *flt = PyNumber_Float(obj);
			if (flt != NULL) {
				v->type = NJS_VALUE_DOUBLE;
				v->u.d = PyFloat_AsDouble(flt);
				Py_DECREF(flt);
			} else {
				/* Not a float? Could be complex..? */
				PyErr_Clear();
				goto special;
			}
		}
	}
	else if (PyString_Check(obj)) {
		njs_type_make_string(interp, v, PyString_AsString(obj),
			PyString_Size(obj));
	}
	else if (PyBuffer_Check(obj)) {
		const char *buffer;
		int buffer_len;
		if (PyObject_AsCharBuffer(obj, &buffer, &buffer_len) == 0)
			njs_type_make_string(interp, v, (char *)buffer, 
			    buffer_len);
		else
			goto special;
	}
	else if (PySequence_Check(obj)) {
		int i;
		int len = PySequence_Size(obj);
		njs_type_make_array(interp, v, len);
		for (i = 0; i < len; i++) {
			PyObject *item = PySequence_GetItem(obj, i);
			if (!PyObject_to_NJSValue(interp, &v->u.array->data[i],
			    item, error_return))
			{
				Py_XDECREF(item);
				return 0;
			}
			Py_XDECREF(item);
		}
	} else {
special:
		/* Unknown python objects are wrapped in the special class */
		Py_INCREF(obj);
		if (!njs_instantiate_class(interp, NjsPyClass, obj,
		    decref_callback, v))
		{
			snprintf(error_return, NJS_ERROR_RETURN_SIZE,
				"njs_instantiate_class");
			Py_DECREF(obj);
			return 0;
		}
	}
	return 1;
}

/*------------------------------------------------------------
 * NJSClass
 */

typedef struct {
	PyObject_HEAD
	NJSClassPtr	cls;
	int		foreign;
	PyObject	*context;
	PyObject	*set_prop;
	PyObject	*get_prop;
	PyObject	*constructor;
	PyObject	*generic_method;
} NjsClassObject;

staticforward PyTypeObject NjsClass_Type;

#define NjsClassObject_Check(v)	((v)->ob_type == &NjsClass_Type)

/* Construct an uninitialised class object */
static NjsClassObject *
newNjsClassObject(void)
{
	NjsClassObject *self;
	self = PyObject_New(NjsClassObject, &NjsClass_Type);
	if (self == NULL)
		return NULL;
	self->cls = NULL;		/* XXX always ask for auto_destroy */
	self->context = NULL;
	self->set_prop = NULL;
	self->get_prop = NULL;
	self->constructor = NULL;
	self->generic_method = NULL;
	self->foreign = 0;
	return self;
}

static void
NjsClass_dealloc(NjsClassObject *self)
{
	Py_XDECREF(self->context);
	Py_XDECREF(self->set_prop);
	Py_XDECREF(self->get_prop);
	Py_XDECREF(self->constructor);
	Py_XDECREF(self->generic_method);
	PyObject_Del(self);
}

/* Utility NJSFreeProc callback to 'destroy' python object contexts */
static void
decref_callback(void *context)
{
	PyObject *obj = (PyObject *)context;
	Py_DECREF(obj);
}

/* Constructor callback for making instances of a class */
static NJSMethodResult
njsclass_constructor_callback(NJSClassPtr cls, NJSInterpPtr interp, 
	int argc, NJSValue *argv, void **ictx_return, 
	NJSFreeProc *ictx_destructor_return, char *error_return)
{
	NjsClassObject *self = (NjsClassObject *)njs_class_context(cls);
	NjsInterpObject *interpobj;
	PyObject *args;
	PyObject *result;
	NJSMethodResult ret;

	if (self->constructor == NULL) {
		snprintf(error_return, NJS_ERROR_RETURN_SIZE, 
			"njsclass_constructor_callback: no constructor");
		return NJS_ERROR;
	}

	Py_INCREF(self);
	interpobj = newNjsInterpObject(interp, 0);
	args = NJSValues_to_PyObjects(interpobj, argc, argv);

	/*
	 * Call the constructor function as f(class, interp, argtuple).
	 * It should return an instance context object.
	 */
	result = PyObject_CallFunction(self->constructor, "OOO",
		self, interpobj, args);

	if (result == NULL) {
		njserror(interp, error_return);
		ret = NJS_ERROR;
	} else {
		*ictx_return = (void *)result;
		*ictx_destructor_return = decref_callback; /* DECREF later */
		ret = NJS_OK;
	}

	Py_XDECREF(args);
	Py_XDECREF(interpobj);
	Py_DECREF(self);
	return ret;
}

/*
 * Generic instance property callback.
 * This is like python's setattr/getattr but performed using the
 * class' "generic" method on the instance context.
 * The python callback function can throw ImmutableError or PrefinedError if it wants.
 */
static NJSGenericResult
njsclass_generic_property_callback(NJSClassPtr cls, void *inst_ctx,
	NJSInterpPtr interp, const char *name, 
	int setp, NJSValue *value, char *error_return)
{
	NjsClassObject *self = (NjsClassObject *)njs_class_context(cls);
	PyObject *instance_context;
	NjsInterpObject *interpobj;
	NJSGenericResult ret;
	PyObject *result;
	PyObject *arg;

	interpobj = newNjsInterpObject(interp, 0);
	if (interpobj == NULL) {
		njserror(interp, error_return);
		return NJS_GR_FAILED;
	}

	if (setp) {
		arg = NJSValue_to_PyObject(interpobj, value);
		if (arg == NULL) {
			njserror(interp, error_return);
			Py_DECREF(interpobj);
			return NJS_GR_FAILED;
		}
	} else {
		arg = Py_None;
		Py_INCREF(arg);
	}

	if (inst_ctx == NULL)
		instance_context = Py_None;
	else
		instance_context = (PyObject *)inst_ctx;
	Py_INCREF(instance_context);

	if (setp)
		result = PyObject_CallFunction(self->set_prop, "OOOzO",
			self, instance_context, interpobj, name, arg);
	else
		result = PyObject_CallFunction(self->get_prop, "OOOz",
			self, instance_context, interpobj, name);

	if (result == NULL) {
		PyObject *err = PyErr_Occurred();
		if (err == NjsPredefinedErrorObject) {
			ret = NJS_GR_PREDEF; 
			PyErr_Clear();
		} else if (err == NjsImmutableErrorObject) {
			ret = NJS_GR_IMMUTABLE;
			PyErr_Clear();
		} else {
			njserror(interp, error_return);
			ret = NJS_GR_FAILED;
		}
	} else {
		ret = NJS_GR_HANDLED;
		if (!setp) {
			if (!PyObject_to_NJSValue(interp, value, result, 
			    error_return))
			{
				ret = NJS_GR_FAILED;
			}
		} else	
			; /* XXX return value from set_prop() ignored */
		Py_DECREF(result);
	}

	Py_DECREF(instance_context);
	Py_DECREF(arg);
	return ret;
}

/* Set the function to be used for the "generic" property method */
static PyObject *
NjsClass_define_generic_property(NjsClassObject *self, PyObject *args)
{
	int flags;
	PyObject *getfn, *setfn;

	if (!PyArg_ParseTuple(args, "iOO:define_generic_property", &flags,
	    &getfn, &setfn))
		return NULL;
	if ((flags & (NJS_CF_STATIC|NJS_CF_IMMUTABLE|NJS_CF_DONTDELETE))
	    != flags)
	{
		PyErr_SetString(PyExc_ValueError, "flags");
		return NULL;
	}
	if (!PyCallable_Check(getfn) || !PyCallable_Check(setfn)) {
		PyErr_SetString(PyExc_TypeError, "callable");
		return NULL;
	} else {
		Py_XDECREF(self->set_prop);
		Py_XDECREF(self->get_prop);
		self->set_prop = setfn;
		self->get_prop = getfn;
		Py_INCREF(setfn);
		Py_INCREF(getfn);
		njs_class_define_generic_property(self->cls, flags, 
			njsclass_generic_property_callback);
	}

	Py_INCREF(Py_None);
	return Py_None;
}

/* Specific instance method callback (NJSGenericMethProc) */
static NJSMethodResult
njsclass_generic_method_callback(NJSClassPtr cls, void *inst_ctx, 
	NJSInterpPtr interp, const char *name,
	int argc, NJSValue *argv, NJSValue *result_return,
	char *error_return)
{
	NjsClassObject *self = (NjsClassObject *)njs_class_context(cls);
	NjsInterpObject *interpobj;
	PyObject *result;
	PyObject *args;
	PyObject *instance_context;
	int ok;

	if (!PyCallable_Check(self->generic_method))  {
		snprintf(error_return, NJS_ERROR_RETURN_SIZE,
			"njsclass_generic_method_callback: not callable");
		return NJS_ERROR;
	}

	interpobj = newNjsInterpObject(interp, 0);
	if (interpobj == NULL) {
		njserror(interp, error_return);
		return NJS_ERROR;
	}

	args = NJSValues_to_PyObjects(interpobj, argc, argv);
	if (args == NULL) {
		njserror(interp, error_return);
		Py_DECREF(interpobj);
		return NJS_ERROR;
	}

	if (inst_ctx == NULL)
		instance_context = Py_None;
	else
		instance_context = (PyObject *)inst_ctx;

	Py_INCREF(instance_context);
	Py_INCREF(self);

	result = PyObject_CallFunction(self->generic_method,
	    "OOOzO", self, interpobj, instance_context, name, args);

	Py_DECREF(instance_context);

	if (result == NULL) {
		njserror(interp, error_return);
		Py_DECREF(self);
		Py_DECREF(args);
		Py_DECREF(interpobj);
		return NJS_ERROR;
	}

	ok = PyObject_to_NJSValue(interp, result_return, result, error_return);

	Py_DECREF(self);
	Py_DECREF(args);
	Py_DECREF(interpobj);
	Py_DECREF(result);

	return ok ? NJS_OK : NJS_ERROR;
}

/* Set the function to be used for a method callback */
static PyObject *
NjsClass_define_generic_method(NjsClassObject *self, PyObject *args)
{
	int flags;
	PyObject *fn;

	if (!PyArg_ParseTuple(args, "iO:define_generic_method", &flags, &fn))
		return NULL;
	if ((flags & (NJS_CF_STATIC|NJS_CF_IMMUTABLE|NJS_CF_DONTDELETE))
	    != flags)
	{
		PyErr_SetString(PyExc_ValueError, "flags");
		return NULL;
	}
	if (!PyCallable_Check(fn)) {
		PyErr_SetString(PyExc_TypeError, "callable");
		return NULL;
	} else {
		Py_XDECREF(self->generic_method);
		self->generic_method = fn;
		Py_INCREF(fn);
		njs_class_define_generic_method(self->cls, flags, 
			njsclass_generic_method_callback);
	}

	Py_INCREF(Py_None);
	return Py_None;
}

/* Construct an instance of the NJS class object */
static PyObject *
NjsClass_instantiate_class(NjsClassObject *self, PyObject *args)
{
	NjsInterpObject *interp;
	PyObject *context = Py_None;
	NJSValue val;

	if (!PyArg_ParseTuple(args, "O!|O:instantiate_class", 
		&NjsInterp_Type, &interp, &context))
		return NULL;
	Py_INCREF(context);
	if (!njs_instantiate_class(interp->interp, self->cls, context,
	    decref_callback, &val))
	{
		Py_DECREF(context);
		return NjsError(interp->interp, "njs_instantiate_class");
	}

	return NJSValue_to_PyObject(interp, &val);
}

static PyMethodDef NjsClass_methods[] = {
	{"define_generic_property",
		(PyCFunction)NjsClass_define_generic_property, METH_VARARGS},
	{"define_generic_method",
		(PyCFunction)NjsClass_define_generic_method, METH_VARARGS},
	{"instantiate_class",
		(PyCFunction)NjsClass_instantiate_class, METH_VARARGS},
	{NULL,							/* sentinel */
		NULL}
};

/* Return a normal python attribute from a NJSClass object */
static PyObject *
NjsClass_getattr(NjsClassObject *self, char *name)
{
	/* context only available if one was set */
	if (strcmp(name, "context") == 0 && self->context != NULL) {
		Py_INCREF(self->context);
		return self->context;
	}
	return Py_FindMethod(NjsClass_methods, (PyObject *)self, name);
}

statichere PyTypeObject NjsClass_Type = {
	/* The ob_type field must be initialized in the module init function
	 * to be portable to Windows without using C++. */
	PyObject_HEAD_INIT(NULL)
	0,				/*ob_size*/
	"NjsClass",			/*tp_name*/
	sizeof(NjsClassObject),		/*tp_basicsize*/
	0,				/*tp_itemsize*/
	/* methods */
	(destructor)NjsClass_dealloc,	/*tp_dealloc*/
	0,				/*tp_print*/
	(getattrfunc)NjsClass_getattr,	/*tp_getattr*/
	(setattrfunc)0,			/*tp_setattr*/
	0,				/*tp_compare*/
	0,				/*tp_repr*/
	0,				/*tp_as_number*/
	0,				/*tp_as_sequence*/
	0,				/*tp_as_mapping*/
	0,				/*tp_hash*/
};

/*------------------------------------------------------------
 * NJSInterp
 */

/* Construct a reference to an existing interpreter */
static NjsInterpObject *
newNjsInterpObject(NJSInterpPtr interp, int destroy_on_deref)
{
	NjsInterpObject *self;
	self = PyObject_New(NjsInterpObject, &NjsInterp_Type);
	if (self == NULL)
		return NULL;
	self->interp = interp;
	self->destroy_on_deref = destroy_on_deref;
	return self;
}

/* Utility function to return the last interpreter result */
static PyObject *
interp_result(NjsInterpObject *self)
{
	NJSValue result;
	njs_result(self->interp, &result);
	return NJSValue_to_PyObject(self, &result);
}

/* Deallocate an interpreter reference */
static void
NjsInterp_dealloc(NjsInterpObject *self)
{
	if (self->destroy_on_deref)
		njs_destroy_interp(self->interp);
	PyObject_Del(self);
}

/* Set a global variable in the interpeter */
static PyObject *
NjsInterp_set_var(NjsInterpObject *self, PyObject *args)
{
	char *name;
	PyObject *value;
	NJSValue v;
	char error[NJS_ERROR_RETURN_SIZE];

	if (!PyArg_ParseTuple(args, "zO:set_var", &name, &value))
		return NULL;
	if (!PyObject_to_NJSValue(self->interp, &v, value, error))
		return NjsError(self->interp, error);
	njs_set_var(self->interp, name, &v);
	Py_INCREF(Py_None);
	return Py_None;
}

/* Retreive a global variable from the interpeter */
static PyObject *
NjsInterp_get_var(NjsInterpObject *self, PyObject *args)
{
	char *name;
	NJSValue v;

	if (!PyArg_ParseTuple(args, "z:get_var", &name))
		return NULL;
	njs_get_var(self->interp, name, &v);
	return NJSValue_to_PyObject(self, &v);
}

/* Return the current interpreter options */
static PyObject *
NjsInterp_get_options(NjsInterpObject *self, PyObject *args)
{
	NjsInterpOptionsObject *opt;
	if (!PyArg_ParseTuple(args, ":get_options"))
		return NULL;
	opt = newNjsInterpOptionsObject();
	if (opt != NULL)
		njs_get_options(self->interp, &opt->value);
	if (opt->value.s_context != NULL) {
		NjsInterpOptionsObject *orig;
		orig = (NjsInterpOptionsObject *)opt->value.s_context;
		if (NjsInterpOptionsObject_Check(orig)) {
			Py_DECREF(opt);
			Py_INCREF(orig);
		}
		opt = orig;
	}
	return (PyObject *)opt;
}

/* Update the current interpreter options */
static PyObject *
NjsInterp_set_options(NjsInterpObject *self, PyObject *args)
{
	NjsInterpOptionsObject *opt;
	if (!PyArg_ParseTuple(args, "O!:set_options",
		&NjsInterpOptions_Type, &opt))
		return NULL;
	njs_set_options(self->interp, &opt->value);
	Py_INCREF(Py_None);
	return Py_None;
}

/* Evaluate javascript code and return the result */
static PyObject *
NjsInterp_eval(NjsInterpObject *self, PyObject *args)
{
	char *code;
	if (!PyArg_ParseTuple(args, "z:eval", &code))
		return NULL;
	if (!njs_eval(self->interp, code))
		return NjsError(self->interp, "eval");
	return interp_result(self);
}

/* Evaluate a JavaScript file and return None */
static PyObject *
NjsInterp_eval_javascript_file(NjsInterpObject *self, PyObject *args)
{
	char *fname;
	if (!PyArg_ParseTuple(args, "z:eval_javascript_file", &fname))
		return NULL;
	if (!njs_eval_javascript_file(self->interp, fname))
		return NjsError(self->interp, "eval_javascript_file");
	Py_INCREF(Py_None);
	return Py_None;
}

/* Evaluate a NJS bytecode file and return None */
static PyObject *
NjsInterp_execute_byte_code_file(NjsInterpObject *self, PyObject *args)
{
	char *fname;
	if (!PyArg_ParseTuple(args, "z:execute_byte_code_file", &fname))
		return NULL;
	if (!njs_execute_byte_code_file(self->interp, fname))
		return NjsError(self->interp, "execute_byte_code_file");
	Py_INCREF(Py_None);
	return Py_None;
}

/* Call the named function with the given arguments */
static PyObject *
NjsInterp_apply(NjsInterpObject *self, PyObject *args)
{
	char *name;
	PyObject *argseq;
	NJSValue argarray;
	char error[NJS_ERROR_RETURN_SIZE];

	if (!PyArg_ParseTuple(args, "zO:apply", &name, &argseq))
		return NULL;
	if (!PySequence_Check(argseq)) {
		PyErr_SetString(PyExc_TypeError, "not sequence");
		return NULL;
	}

	if (!PyObject_to_NJSValue(self->interp, &argarray, argseq, error))
		return NjsError(self->interp, error);
	/* XXX argseq may be GCd while it is not on the apply stack? */
	if (!njs_apply(self->interp, name, 
	    argarray.u.array->length, argarray.u.array->data))
		return NjsError(self->interp, "apply");
	return interp_result(self);
}

/* Compile the given input file into an asm file and/or a bytecode file */
static PyObject *
NjsInterp_compile(NjsInterpObject *self, PyObject *args)
{
	char *input_file;
	char *assembler_file = NULL;
	char *byte_code_file = NULL;

	if (!PyArg_ParseTuple(args, "z|zz:compile", &input_file, 
	    &assembler_file, &byte_code_file))
		return NULL;
	if (!njs_compile(self->interp, input_file, assembler_file, 
	    byte_code_file))
		return NjsError(self->interp, "compile");
	Py_INCREF(Py_None);
	return Py_None;
}

/* Compile a file into a bytecode string */
static PyObject *
NjsInterp_compile_to_byte_code(NjsInterpObject *self, PyObject *args)
{
	char *input_file;
	unsigned char *buf;
	unsigned int buflen;

	if (!PyArg_ParseTuple(args, "z:compile_to_byte_code", &input_file))
		return NULL;
	if (!njs_compile_to_byte_code(self->interp, input_file, &buf, &buflen))
		return NjsError(self->interp, "compile_to_byte_code");
	return PyString_FromStringAndSize((const char *)buf, buflen);
}

/* Compile a program string into a bytecode string */
static PyObject *
NjsInterp_compile_data_to_byte_code(NjsInterpObject *self, PyObject *args)
{
	char *data;
	int datalen;
	unsigned char *buf;
	unsigned int buflen;

	if (!PyArg_ParseTuple(args, "z#:compile_data_to_byte_code", 
	    &data, &datalen))
		return NULL;

	if (!njs_compile_data_to_byte_code(self->interp, data, datalen,
	    &buf, &buflen))
		return NjsError(self->interp, "compile_to_byte_code");
	return PyString_FromStringAndSize((const char *)buf, buflen);
}

/* Execute a precompiled bytecode string */
static PyObject *
NjsInterp_execute_byte_code(NjsInterpObject *self, PyObject *args)
{
	char *bc;
	size_t bclen;
	if (!PyArg_ParseTuple(args, "t#:execute_byte_code", &bc, &bclen))
		return NULL;
	if (!njs_execute_byte_code(self->interp, bc, bclen))
		return NjsError(self->interp, "execute_byte_code");
	Py_INCREF(Py_None);
	return Py_None;
}

/* 
 * Global method callback.
 * The arguments to the JS method are converted to a Python tuple,
 * and then the python function is invoked as f(interp, argtuple).
 */
static NJSMethodResult
global_method_callback(void *context, NJSInterpPtr interp, int argc, 
	NJSValue *argv, NJSValue *result_return, char *error_return)
{
	PyObject *func = (PyObject *)context;
	NjsInterpObject *interpobj = newNjsInterpObject(interp, 0);
	PyObject *args = NJSValues_to_PyObjects(interpobj, argc, argv);
	NJSMethodResult ret;
	PyObject *result;

	result = PyObject_CallFunction(func, "OO", interpobj, args);
	if (result == NULL) {
		njserror(interp, error_return);
		ret = NJS_ERROR;
	} else {
		if (!PyObject_to_NJSValue(interp, result_return, result, 
		    error_return))
			ret = NJS_ERROR;
		else
			ret = NJS_OK;
	}
	Py_DECREF(result);
	Py_DECREF(interpobj);
	Py_DECREF(args);
	return ret;
}

/*
 * Create a global JS method in this interpreter.
 * When the method is called, it will redirect to the given Python function.
 */
static PyObject *
NjsInterp_create_global_method(NjsInterpObject *self, PyObject *args)
{
	char *name;
	PyObject *func;
	if (!PyArg_ParseTuple(args, "zO:create_global_method", &name, &func))
		return NULL;
	if (!PyCallable_Check(func)) {
		PyErr_SetString(PyExc_TypeError, "not callable");
		return NULL;
	}
	Py_INCREF(func);
	if (!njs_create_global_method(self->interp, name, 
	    global_method_callback, (void *)func, decref_callback))
	{
		Py_DECREF(func);
		return NjsError(self->interp, "create_global_method");
	}
	Py_INCREF(Py_None);
	return Py_None;
}

/* Insert a class object into the interpreter's global namespace */
/* XXX not documented, but this really only works for the one interpreter */
static PyObject *
NjsInterp_define_class(NjsInterpObject *self, PyObject *args)
{
	NjsClassObject *cls;
	char *name;

	if (!PyArg_ParseTuple(args, "O!z:define_class",
	    &NjsClass_Type, &cls, &name))
		return NULL;
	if (!njs_define_class(self->interp, cls->cls, name))
		return NjsError(self->interp, "define_class");
	Py_INCREF(Py_None);
	return Py_None;
}

/* Find a class by its name in this interpreter's global namespace */
static PyObject *
NjsInterp_lookup_class(NjsInterpObject *self, PyObject *args)
{
	char *name;
	NjsClassObject *cls;
	NJSClassPtr clsptr;

	if (!PyArg_ParseTuple(args, "z:lookup_class", &name))
		return NULL;
	clsptr = njs_lookup_class(self->interp, name);
	if (clsptr == NULL) {
		PyErr_SetString(PyExc_KeyError, name);
		return NULL;
	}
	cls = newNjsClassObject();
	cls->foreign = 1;
	cls->cls = clsptr;
	return (PyObject *)cls;
}

/* Return instance context in a tuple if the first argument is an 
 * instance of the 2nd class arg; else return None */
static PyObject *
NjsInterp_isa(NjsInterpObject *self, PyObject *args)
{
	NjsClassObject *cls;
	PyObject *obj;
	NJSValue val;
	void *ctxt = NULL;
	char error[NJS_ERROR_RETURN_SIZE];

	if (!PyArg_ParseTuple(args, "OO!:isa", &obj, &NjsClass_Type, &cls))
		return NULL;
	if (!PyObject_to_NJSValue(self->interp, &val, obj, error))
		return NjsError(self->interp, error);
	if (!njs_isa(self->interp, &val, cls->cls, &ctxt)) {
		Py_INCREF(Py_None);
		return Py_None;
	}
	/* Can't risk having foreign context pointers cast to PyObject */
	if (cls->foreign)
		return PyInt_FromLong(1);
	return Py_BuildValue("(O)", ctxt ? (PyObject *)ctxt : Py_None);
}

/* Add directory to this interpreter's extension path */
static PyObject *
NjsInterp_ext_add_directory(NjsInterpObject *self, PyObject *args)
{
	char *dir;

	if (!PyArg_ParseTuple(args, "z:ext_add_directory", &dir))
		return NULL;
	if (!njs_ext_add_directory(self->interp, dir))
		return NjsError(self->interp, "ext_add_directory");
	Py_INCREF(Py_None);
	return Py_None;
}

/* Remove directory from this interpreter's extension path */
static PyObject *
NjsInterp_ext_remove_directory(NjsInterpObject *self, PyObject *args)
{
	char *dir;

	if (!PyArg_ParseTuple(args, "z:ext_remove_directory", &dir))
		return NULL;
	if (!njs_ext_remove_directory(self->interp, dir))
		return NjsError(self->interp, "ext_remove_directory");
	Py_INCREF(Py_None);
	return Py_None;
}

/* Reset this interpreter's extension path to its default */
static PyObject *
NjsInterp_ext_default_directories(NjsInterpObject *self, PyObject *args)
{
	if (!PyArg_ParseTuple(args, ":ext_default_directories"))
		return NULL;
	if (!njs_ext_default_directories(self->interp))
		return NjsError(self->interp, "ext_default_directories");
	Py_INCREF(Py_None);
	return Py_None;
}

/* Search for and load into this interpreter an extension by name */
static PyObject *
NjsInterp_ext_load_module(NjsInterpObject *self, PyObject *args)
{
	char *name;

	if (!PyArg_ParseTuple(args, "z:ext_load_module", &name))
		return NULL;
	if (!njs_ext_load_module(self->interp, name))
		return NjsError(self->interp, "ext_load_module");
	Py_INCREF(Py_None);
	return Py_None;
}

static PyMethodDef NjsInterp_methods[] = {
	{"set_var",
		(PyCFunction)NjsInterp_set_var,			METH_VARARGS},
	{"get_var",
		(PyCFunction)NjsInterp_get_var,			METH_VARARGS},
	{"get_options",
		(PyCFunction)NjsInterp_get_options,		METH_VARARGS},
	{"set_options",
		(PyCFunction)NjsInterp_set_options,		METH_VARARGS},
	{"eval",
		(PyCFunction)NjsInterp_eval,			METH_VARARGS},
	{"eval_javascript_file",
		(PyCFunction)NjsInterp_eval_javascript_file,	METH_VARARGS},
	{"execute_byte_code_file",
		(PyCFunction)NjsInterp_execute_byte_code_file,	METH_VARARGS},
	{"apply",
		(PyCFunction)NjsInterp_apply,			METH_VARARGS},
	{"compile",
		(PyCFunction)NjsInterp_compile,			METH_VARARGS},
	{"compile_to_byte_code",
		(PyCFunction)NjsInterp_compile_to_byte_code,	METH_VARARGS},
	{"compile_data_to_byte_code",
		(PyCFunction)NjsInterp_compile_data_to_byte_code, METH_VARARGS},
	{"execute_byte_code",
		(PyCFunction)NjsInterp_execute_byte_code,	METH_VARARGS},
	{"create_global_method",
		(PyCFunction)NjsInterp_create_global_method,	METH_VARARGS},
	{"define_class",
		(PyCFunction)NjsInterp_define_class,		METH_VARARGS},
	{"lookup_class",
		(PyCFunction)NjsInterp_lookup_class,		METH_VARARGS},
	{"isa",
		(PyCFunction)NjsInterp_isa,			METH_VARARGS},
	{"ext_add_directory",
		(PyCFunction)NjsInterp_ext_add_directory,	METH_VARARGS},
	{"ext_remove_directory",
		(PyCFunction)NjsInterp_ext_remove_directory,	METH_VARARGS},
	{"ext_default_directories",
		(PyCFunction)NjsInterp_ext_default_directories,	METH_VARARGS},
	{"ext_load_module",
		(PyCFunction)NjsInterp_ext_load_module,		METH_VARARGS},
	{NULL, NULL} /* sentinel */
};

/* Return an interpreter attribute by name */
static PyObject *
NjsInterp_getattr(NjsInterpObject *self, char *name)
{
	if (strcmp(name, "error_message") == 0)
		return PyString_FromString(njs_error_message(self->interp));

	if (strcmp(name, "result") == 0) {
		NJSValue result;
		njs_result(self->interp, &result);
		return NJSValue_to_PyObject(self, &result);
	}

	return Py_FindMethod(NjsInterp_methods, (PyObject *)self, name);
}

/* Don't allow the setting of any attribute */
static int
NjsInterp_setattr(NjsInterpObject *self, char *name, PyObject *v)
{
	PyErr_SetString(PyExc_AttributeError, name);
	return -1;
}

statichere PyTypeObject NjsInterp_Type = {
	/* The ob_type field must be initialized in the module init function
	 * to be portable to Windows without using C++. */
	PyObject_HEAD_INIT(NULL)
	0,			/*ob_size*/
	"NjsInterp",			/*tp_name*/
	sizeof(NjsInterpObject),	/*tp_basicsize*/
	0,			/*tp_itemsize*/
	/* methods */
	(destructor)NjsInterp_dealloc, /*tp_dealloc*/
	0,			/*tp_print*/
	(getattrfunc)NjsInterp_getattr, /*tp_getattr*/
	(setattrfunc)NjsInterp_setattr, /*tp_setattr*/
	0,			/*tp_compare*/
	0,			/*tp_repr*/
	0,			/*tp_as_number*/
	0,			/*tp_as_sequence*/
	0,			/*tp_as_mapping*/
	0,			/*tp_hash*/
};

/*------------------------------------------------------------
 * njs module 
 */

/* Create a default options object (not bound to any interpreter) */
static PyObject *
Njs_create_default_options(PyObject *self, PyObject *args)
{
	NjsInterpOptionsObject *opt;

	if (!PyArg_ParseTuple(args, ":create_default_options"))
		return NULL;
	opt = newNjsInterpOptionsObject();
	njs_init_default_options(&opt->value);
	opt->value.s_context = opt;
	opt->value.hook_context = opt;

	return (PyObject *)opt;
}

/* Create a new interpreter object */
static PyObject *
Njs_create_interp(PyObject *self, PyObject *args)
{
	NjsInterpObject *result;
	NJSInterpPtr interp;
	NjsInterpOptionsObject *opt = NULL;

	if (!PyArg_ParseTuple(args, "|O!:create_interp",
		&NjsInterpOptions_Type, &opt))
		return NULL;

	if (opt == NULL) {
		NJSInterpOptions options;
		njs_init_default_options(&options);
		interp = njs_create_interp(&options);
	} else
		interp = njs_create_interp(&opt->value);

	if (interp == NULL) {
		PyErr_SetString(PyExc_RuntimeError, "njs_create_interp");
		return NULL;
	}

	result = newNjsInterpObject(interp, 1);
	return (PyObject *)result;
}

/* Create a new class (not bound to any interpreter) */
static PyObject *
Njs_create_class(PyObject *self, PyObject *args)
{
	PyObject *context = Py_None;
	PyObject *constructor = NULL;
	NjsClassObject *res;

	if (!PyArg_ParseTuple(args, "|OO:create_class",
		&context, &constructor))
		return NULL;

	if (constructor != NULL && !PyCallable_Check(constructor)) {
		PyErr_SetString(PyExc_TypeError, "constructor not callable");
		return NULL;
	}

	/*
	 * This is done a little backwards because we need the 
	 * PyObject for the class context before we can get the 
	 * class reference (because there is no API to change a 
	 * class's context.)
	 */
	res = newNjsClassObject();
	if (res == NULL)
		return NULL;

	res->context = context;
	Py_INCREF(context);
	if (constructor != NULL) {
		res->constructor = constructor;
		Py_INCREF(constructor);
	}

	Py_INCREF(res);	/* for decref_callback */
	res->cls = njs_class_create(res, decref_callback, 0, 
		constructor == NULL ? NULL : njsclass_constructor_callback);

	return (PyObject *)res;
}


static PyMethodDef Njs_methods[] = {
	{"create_default_options",
		Njs_create_default_options,		METH_VARARGS},
	{"create_interp",
		Njs_create_interp,			METH_VARARGS},
	{"create_class",
		Njs_create_class,			METH_VARARGS},
	{NULL,		NULL}		/* sentinel */
};

/* Constructor callback indicating construction not allowed. */
static NJSMethodResult
bad_constructor_callback(NJSClassPtr cls, NJSInterpPtr interp, 
	int argc, NJSValue *argv, void **ictx_return, 
	NJSFreeProc *ictx_destructor_return, char *error_return)
{
	snprintf(error_return, NJS_ERROR_RETURN_SIZE,
		"cannot construct Python objects");
	return NJS_ERROR;
}

/* Initialise the njs module */

DL_EXPORT(void)
initnjs(void)
{
	PyObject *m, *d;

	/* Initialize the type of the new type object here; doing it here
	 * is required for portability to Windows without requiring C++. */
	NjsBoolean_Type.ob_type	= &PyType_Type;
	NjsUndefined_Type.ob_type = &PyType_Type;
	NjsVMDispatchMethod_Type.ob_type = &PyType_Type;
	NjsInterpOptions_Type.ob_type = &PyType_Type;
	NjsValue_Type.ob_type = &PyType_Type;
	NjsClass_Type.ob_type = &PyType_Type;
	NjsInterp_Type.ob_type = &PyType_Type;

	/* Register a 'Python' object class with NJS */
	NjsPyClass = njs_class_create(NULL, NULL, 1, bad_constructor_callback);

	/* Create the module and add the functions */
	m = Py_InitModule("njs", Njs_methods);

	/* The generic error object */
	NjsErrorObject = PyErr_NewException("njs.Error", NULL, NULL);
	NjsImmutableErrorObject = PyErr_NewException("njs.ImmutableError", NULL, NULL);
	NjsPredefinedErrorObject = PyErr_NewException("njs.PredefinedError", NULL, NULL);
	NjsTrueObject = (PyObject *)newNjsBooleanObject(1);
	NjsFalseObject = (PyObject *)newNjsBooleanObject(0);
	NjsUndefinedObject = newNjsUndefinedObject();

	/* Add some symbolic constants to the module */
	d = PyModule_GetDict(m);

	/* Unique constants */
	PyDict_SetItemString(d, "Error", NjsErrorObject);
	PyDict_SetItemString(d, "ImmutableError", NjsImmutableErrorObject);
	PyDict_SetItemString(d, "PredefinedError", NjsPredefinedErrorObject);
	PyDict_SetItemString(d, "true", NjsTrueObject);
	PyDict_SetItemString(d, "false", NjsFalseObject);
	PyDict_SetItemString(d, "undefined", NjsUndefinedObject);

	/* Enumerated types */
#define D(d,t) \
	PyDict_SetItemString(d, #t, \
		(PyObject *)newNjsVMDispatchMethodObject(NJS_##t))
	D(d,VM_DISPATCH_SWITCH_BASIC);
	D(d,VM_DISPATCH_SWITCH);
	D(d,VM_DISPATCH_JUMPS);
#undef D

	/* Integer constants */
	PyDict_SetItemString(d, "EVENT_OPERAND_COUNT",
		PyInt_FromLong(NJS_EVENT_OPERAND_COUNT));
	PyDict_SetItemString(d, "EVENT_GARBAGE_COLLECT",
		PyInt_FromLong(NJS_EVENT_GARBAGE_COLLECT));

	/* Callable function flags */
	PyDict_SetItemString(d, "CF_STATIC",
		PyInt_FromLong(NJS_CF_STATIC));
	PyDict_SetItemString(d, "CF_IMMUTABLE",
		PyInt_FromLong(NJS_CF_IMMUTABLE));
	PyDict_SetItemString(d, "CF_DONTDELETE",
		PyInt_FromLong(NJS_CF_DONTDELETE));

	/* The version number of the underlying NJS library */
	PyDict_SetItemString(d, "version",
		PyString_FromString(njs_version()));

	/* The version number of this module */
	PyDict_SetItemString(d, "__version__",
		PyString_FromString(VERSION));

}
